@@ -24,6 +24,7 @@
 #include "tcpiohandler-mplexer.hh"
 #include "dnswriter.hh"
 #include "dolog.hh"
+#include "dnsdist-random.hh"
 #include "dnsdist-tcp.hh"
 #include "dnsdist-nghttp2.hh"
 #include "dnsdist-session-cache.hh"
@@ -54,64 +55,6 @@ struct HealthCheckData
   bool d_initial{false};
 };
 
-void updateHealthCheckResult(const std::shared_ptr<DownstreamState>& dss, bool initial, bool newState)
-{
-  if (initial) {
-    warnlog("Marking downstream %s as '%s'", dss->getNameWithAddr(), newState ? "up" : "down");
-    dss->setUpStatus(newState);
-    return;
-  }
-
-  if (newState) {
-    /* check succeeded */
-    dss->currentCheckFailures = 0;
-
-    if (!dss->upStatus) {
-      /* we were marked as down */
-      dss->consecutiveSuccessfulChecks++;
-      if (dss->consecutiveSuccessfulChecks < dss->minRiseSuccesses) {
-        /* if we need more than one successful check to rise
-           and we didn't reach the threshold yet,
-           let's stay down */
-        newState = false;
-      }
-    }
-  }
-  else {
-    /* check failed */
-    dss->consecutiveSuccessfulChecks = 0;
-
-    if (dss->upStatus) {
-      /* we are currently up */
-      dss->currentCheckFailures++;
-      if (dss->currentCheckFailures < dss->maxCheckFailures) {
-        /* we need more than one failure to be marked as down,
-           and we did not reach the threshold yet, let's stay up */
-        newState = true;
-      }
-    }
-  }
-
-  if (newState != dss->upStatus) {
-    warnlog("Marking downstream %s as '%s'", dss->getNameWithAddr(), newState ? "up" : "down");
-
-    if (newState && !dss->isTCPOnly() && (!dss->connected || dss->reconnectOnUp)) {
-      newState = dss->reconnect();
-
-      if (dss->connected && !dss->threadStarted.test_and_set()) {
-        dss->tid = std::thread(responderThread, dss);
-      }
-    }
-
-    dss->setUpStatus(newState);
-    dss->currentCheckFailures = 0;
-    dss->consecutiveSuccessfulChecks = 0;
-    if (g_snmpAgent && g_snmpTrapsEnabled) {
-      g_snmpAgent->sendBackendStatusChangeTrap(dss);
-    }
-  }
-}
-
 static bool handleResponse(std::shared_ptr<HealthCheckData>& data)
 {
   auto& ds = data->d_ds;
@@ -145,7 +88,7 @@ static bool handleResponse(std::shared_ptr<HealthCheckData>& data)
       return false;
     }
 
-    if (ds->mustResolve && (responseHeader->rcode == RCode::NXDomain || responseHeader->rcode == RCode::Refused)) {
+    if (ds->d_config.mustResolve && (responseHeader->rcode == RCode::NXDomain || responseHeader->rcode == RCode::Refused)) {
       if (g_verboseHealthChecks) {
         infolog("Backend %s responded to health check with %s while mustResolve is set", ds->getNameWithAddr(), responseHeader->rcode == RCode::NXDomain ? "NXDomain" : "Refused");
       }
@@ -197,15 +140,10 @@ public:
     return true;
   }
 
-  const ClientState* getClientState() const override
-  {
-    return nullptr;
-  }
-
   void handleResponse(const struct timeval& now, TCPResponse&& response) override
   {
     d_data->d_buffer = std::move(response.d_buffer);
-    updateHealthCheckResult(d_data->d_ds, d_data->d_initial, ::handleResponse(d_data));
+    d_data->d_ds->submitHealthCheckResult(d_data->d_initial, ::handleResponse(d_data));
   }
 
   void handleXFRResponse(const struct timeval& now, TCPResponse&& response) override
@@ -213,9 +151,9 @@ public:
     throw std::runtime_error("Unexpected XFR reponse to a health check query");
   }
 
-  void notifyIOError(IDState&& query, const struct timeval& now) override
+  void notifyIOError(InternalQueryState&& query, const struct timeval& now) override
   {
-    updateHealthCheckResult(d_data->d_ds, d_data->d_initial, false);
+    d_data->d_ds->submitHealthCheckResult(d_data->d_initial, false);
   }
 
 private:
@@ -228,26 +166,28 @@ static void healthCheckUDPCallback(int fd, FDMultiplexer::funcparam_t& param)
   data->d_mplexer.removeReadFD(fd);
 
   ComboAddress from;
-  from.sin4.sin_family = data->d_ds->remote.sin4.sin_family;
+  from.sin4.sin_family = data->d_ds->d_config.remote.sin4.sin_family;
   auto fromlen = from.getSocklen();
   data->d_buffer.resize(512);
   auto got = recvfrom(data->d_udpSocket.getHandle(), &data->d_buffer.at(0), data->d_buffer.size(), 0, reinterpret_cast<sockaddr *>(&from), &fromlen);
   if (got < 0) {
     if (g_verboseHealthChecks) {
-      infolog("Error receiving health check response from %s: %s", data->d_ds->remote.toStringWithPort(), stringerror());
+      infolog("Error receiving health check response from %s: %s", data->d_ds->d_config.remote.toStringWithPort(), stringerror());
     }
-    updateHealthCheckResult(data->d_ds, data->d_initial, false);
+    data->d_ds->submitHealthCheckResult(data->d_initial, false);
+    return;
   }
 
   /* we are using a connected socket but hey.. */
-  if (from != data->d_ds->remote) {
+  if (from != data->d_ds->d_config.remote) {
     if (g_verboseHealthChecks) {
-      infolog("Invalid health check response received from %s, expecting one from %s", from.toStringWithPort(), data->d_ds->remote.toStringWithPort());
+      infolog("Invalid health check response received from %s, expecting one from %s", from.toStringWithPort(), data->d_ds->d_config.remote.toStringWithPort());
     }
-    updateHealthCheckResult(data->d_ds, data->d_initial, false);
+    data->d_ds->submitHealthCheckResult(data->d_initial, false);
+    return;
   }
 
-  updateHealthCheckResult(data->d_ds, data->d_initial, handleResponse(data));
+  data->d_ds->submitHealthCheckResult(data->d_initial, handleResponse(data));
 }
 
 static void healthCheckTCPCallback(int fd, FDMultiplexer::funcparam_t& param)
@@ -281,7 +221,7 @@ static void healthCheckTCPCallback(int fd, FDMultiplexer::funcparam_t& param)
     if (data->d_tcpState == HealthCheckData::TCPState::ReadingResponse) {
       ioState = data->d_tcpHandler->tryRead(data->d_buffer, data->d_bufferPos, data->d_buffer.size());
       if (ioState == IOState::Done) {
-        updateHealthCheckResult(data->d_ds, data->d_initial, handleResponse(data));
+        data->d_ds->submitHealthCheckResult(data->d_initial, handleResponse(data));
       }
     }
 
@@ -308,13 +248,13 @@ static void healthCheckTCPCallback(int fd, FDMultiplexer::funcparam_t& param)
     ioGuard.release();
   }
   catch (const std::exception& e) {
-    updateHealthCheckResult(data->d_ds, data->d_initial, false);
+    data->d_ds->submitHealthCheckResult(data->d_initial, false);
     if (g_verboseHealthChecks) {
       infolog("Error checking the health of backend %s: %s", data->d_ds->getNameWithAddr(), e.what());
     }
   }
   catch (...) {
-    updateHealthCheckResult(data->d_ds, data->d_initial, false);
+    data->d_ds->submitHealthCheckResult(data->d_initial, false);
     if (g_verboseHealthChecks) {
       infolog("Unknown exception while checking the health of backend %s", data->d_ds->getNameWithAddr());
     }
@@ -325,10 +265,10 @@ bool queueHealthCheck(std::unique_ptr<FDMultiplexer>& mplexer, const std::shared
 {
   try
   {
-    uint16_t queryID = getRandomDNSID();
-    DNSName checkName = ds->checkName;
-    uint16_t checkType = ds->checkType.getCode();
-    uint16_t checkClass = ds->checkClass;
+    uint16_t queryID = dnsdist::getRandomDNSID();
+    DNSName checkName = ds->d_config.checkName;
+    uint16_t checkType = ds->d_config.checkType.getCode();
+    uint16_t checkClass = ds->d_config.checkClass;
     dnsheader checkHeader;
     memset(&checkHeader, 0, sizeof(checkHeader));
 
@@ -336,13 +276,13 @@ bool queueHealthCheck(std::unique_ptr<FDMultiplexer>& mplexer, const std::shared
     checkHeader.id = queryID;
 
     checkHeader.rd = true;
-    if (ds->setCD) {
+    if (ds->d_config.setCD) {
       checkHeader.cd = true;
     }
 
-    if (ds->checkFunction) {
+    if (ds->d_config.checkFunction) {
       auto lock = g_lua.lock();
-      auto ret = ds->checkFunction(checkName, checkType, checkClass, &checkHeader);
+      auto ret = ds->d_config.checkFunction(checkName, checkType, checkClass, &checkHeader);
       checkName = std::get<0>(ret);
       checkType = std::get<1>(ret);
       checkClass = std::get<2>(ret);
@@ -357,7 +297,7 @@ bool queueHealthCheck(std::unique_ptr<FDMultiplexer>& mplexer, const std::shared
     uint16_t packetSize = packet.size();
     std::string proxyProtocolPayload;
     size_t proxyProtocolPayloadSize = 0;
-    if (ds->useProxyProtocol) {
+    if (ds->d_config.useProxyProtocol) {
       proxyProtocolPayload = makeLocalProxyHeader();
       proxyProtocolPayloadSize = proxyProtocolPayload.size();
       if (!ds->isDoH()) {
@@ -365,41 +305,39 @@ bool queueHealthCheck(std::unique_ptr<FDMultiplexer>& mplexer, const std::shared
       }
     }
 
-    Socket sock(ds->remote.sin4.sin_family, ds->doHealthcheckOverTCP() ? SOCK_STREAM : SOCK_DGRAM);
+    Socket sock(ds->d_config.remote.sin4.sin_family, ds->doHealthcheckOverTCP() ? SOCK_STREAM : SOCK_DGRAM);
 
     sock.setNonBlocking();
-    if (!IsAnyAddress(ds->sourceAddr)) {
+
+#ifdef SO_BINDTODEVICE
+    if (!ds->d_config.sourceItfName.empty()) {
+      int res = setsockopt(sock.getHandle(), SOL_SOCKET, SO_BINDTODEVICE, ds->d_config.sourceItfName.c_str(), ds->d_config.sourceItfName.length());
+      if (res != 0 && g_verboseHealthChecks) {
+        infolog("Error setting SO_BINDTODEVICE on the health check socket for backend '%s': %s", ds->getNameWithAddr(), stringerror());
+      }
+    }
+#endif
+
+    if (!IsAnyAddress(ds->d_config.sourceAddr)) {
       sock.setReuseAddr();
 #ifdef IP_BIND_ADDRESS_NO_PORT
-      if (ds->ipBindAddrNoPort) {
+      if (ds->d_config.ipBindAddrNoPort) {
         SSetsockopt(sock.getHandle(), SOL_IP, IP_BIND_ADDRESS_NO_PORT, 1);
       }
 #endif
-
-      if (!ds->sourceItfName.empty()) {
-#ifdef SO_BINDTODEVICE
-        int res = setsockopt(sock.getHandle(), SOL_SOCKET, SO_BINDTODEVICE, ds->sourceItfName.c_str(), ds->sourceItfName.length());
-        if (res != 0 && g_verboseHealthChecks) {
-          infolog("Error setting SO_BINDTODEVICE on the health check socket for backend '%s': %s", ds->getNameWithAddr(), stringerror());
-        }
-#endif
-      }
-      sock.bind(ds->sourceAddr);
+      sock.bind(ds->d_config.sourceAddr);
     }
 
     auto data = std::make_shared<HealthCheckData>(*mplexer, ds, std::move(checkName), checkType, checkClass, queryID);
     data->d_initial = initialCheck;
 
     gettimeofday(&data->d_ttd, nullptr);
-    data->d_ttd.tv_sec += ds->checkTimeout / 1000; /* ms to seconds */
-    data->d_ttd.tv_usec += (ds->checkTimeout % 1000) * 1000; /* remaining ms to us */
-    if (data->d_ttd.tv_usec > 1000000) {
-      ++data->d_ttd.tv_sec;
-      data->d_ttd.tv_usec -= 1000000;
-    }
+    data->d_ttd.tv_sec += ds->d_config.checkTimeout / 1000; /* ms to seconds */
+    data->d_ttd.tv_usec += (ds->d_config.checkTimeout % 1000) * 1000; /* remaining ms to us */
+    normalizeTV(data->d_ttd);
 
     if (!ds->doHealthcheckOverTCP()) {
-      sock.connect(ds->remote);
+      sock.connect(ds->d_config.remote);
       data->d_udpSocket = std::move(sock);
       ssize_t sent = udpClientSendRequestToBackend(ds, data->d_udpSocket.getHandle(), packet, true);
       if (sent < 0) {
@@ -413,16 +351,16 @@ bool queueHealthCheck(std::unique_ptr<FDMultiplexer>& mplexer, const std::shared
       mplexer->addReadFD(data->d_udpSocket.getHandle(), &healthCheckUDPCallback, data, &data->d_ttd);
     }
     else if (ds->isDoH()) {
-      InternalQuery query(std::move(packet), IDState());
+      InternalQuery query(std::move(packet), InternalQueryState());
       query.d_proxyProtocolPayload = std::move(proxyProtocolPayload);
       auto sender = std::shared_ptr<TCPQuerySender>(new HealthCheckQuerySender(data));
       if (!sendH2Query(ds, mplexer, sender, std::move(query), true)) {
-        updateHealthCheckResult(data->d_ds, data->d_initial, false);
+        data->d_ds->submitHealthCheckResult(data->d_initial, false);
       }
     }
     else {
       time_t now = time(nullptr);
-      data->d_tcpHandler = std::make_unique<TCPIOHandler>(ds->d_tlsSubjectName, sock.releaseHandle(), timeval{ds->checkTimeout,0}, ds->d_tlsCtx, now);
+      data->d_tcpHandler = std::make_unique<TCPIOHandler>(ds->d_config.d_tlsSubjectName, ds->d_config.d_tlsSubjectIsAddr, sock.releaseHandle(), timeval{ds->d_config.checkTimeout,0}, ds->d_tlsCtx);
       data->d_ioState = std::make_unique<IOStateHandler>(*mplexer, data->d_tcpHandler->getDescriptor());
       if (ds->d_tlsCtx) {
         try {
@@ -435,7 +373,7 @@ bool queueHealthCheck(std::unique_ptr<FDMultiplexer>& mplexer, const std::shared
           vinfolog("Unable to restore a TLS session for the DoT healthcheck: %s", e.what());
         }
       }
-      data->d_tcpHandler->tryConnect(ds->tcpFastOpen, ds->remote);
+      data->d_tcpHandler->tryConnect(ds->d_config.tcpFastOpen, ds->d_config.remote);
 
       const uint8_t sizeBytes[] = { static_cast<uint8_t>(packetSize / 256), static_cast<uint8_t>(packetSize % 256) };
       packet.insert(packet.begin() + proxyProtocolPayloadSize, sizeBytes, sizeBytes + 2);
@@ -502,7 +440,7 @@ void handleQueuedHealthChecks(FDMultiplexer& mplexer, bool initial)
           infolog("Timeout while waiting for the health check response from backend %s", data->d_ds->getNameWithAddr());
         }
 
-        updateHealthCheckResult(data->d_ds, initial, false);
+        data->d_ds->submitHealthCheckResult(initial, false);
       }
       catch (const std::exception& e) {
         if (g_verboseHealthChecks) {
@@ -528,7 +466,7 @@ void handleQueuedHealthChecks(FDMultiplexer& mplexer, bool initial)
           infolog("Timeout while waiting for the health check response from backend %s", data->d_ds->getNameWithAddr());
         }
 
-        updateHealthCheckResult(data->d_ds, initial, false);
+        data->d_ds->submitHealthCheckResult(initial, false);
       }
       catch (const std::exception& e) {
         if (g_verboseHealthChecks) {
